/**************************************************************************************************

  Development Board: EasyMx PRO v7 for STM32 ARM
  Microcontroller: STM32F407VGT6
  Oscillator: external 25 MHz crystal
  MPU Frequency: 120 MHz from PLL

  This UART module communicates with the FTDI Serial-to-USB FT232RL chip located on the board.

  PD5  Pin 86 (USART2_TX)
  PD6  Pin 87 (USART2_RX)

  For SYSCLK = 120 MHz and APB1 Prescaler = 4 the peripheral clock for APB1 = 30 MHz
  USART2 is on APB1 BUS and the peripheral clock is 30 MHz
  USARTDIV = Fck / 16 * baud = 30,000,000 / (16 * 9600) = 0d195.3125
  DIV_Mantissa = 0d195 = 0xC3
  DIV_Fraction = 16 * 0d0.3125 = 0d5 or 0x5
  Concatenate DIV Mantissa and DIV_Fraction,
  Therefore USART2_BRR = 0xC35
*/
/**************************************************************************************************
  UART2ForFreeRTOS.c
**************************************************************************************************/

// Include Files
#include <stdint.h>
#include <stddef.h>
#include "FreeRTOSConfig.h"
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"
#include "timers.h"
#include "projdefs.h"
#include "portmacro.h"

#include "UART2_driver.h"
#include "Buffers.h"
#include "USB_Driver3.h"
//#include "MyTimers.h"
#include "RTC.h"


// Variables





// Queues
QueueHandle_t U2RcvByteQueueHandle;     // Queue of receive bytes
QueueHandle_t U2XmtPointerQueue;        // Queue of transmit message pointers

//QueueHandle_t U2RcvMsgQueueHandle;
//QueueHandle_t U2XmtByteQueueHandle;

// Semaphores
SemaphoreHandle_t U2XmtSemHandle;       // Semaphore indicating transmitter is idle



// Buffers
char U2RcvBuf[256];      // Receive messages are assembled in this buffer
char U2RcvBufPtr = 0;    // Pointer to the next free byte in RcvBuf
char *U2XmtBuf;          // Pointer to the current message being transmitted
char U2XmtBufPtr;        // Points to a byte in transmit buffer

// The receive message routine has 2 states
tPointerFcn U2_RxFuncPtr = UART2_Rx1;  // Start in State 1


/**************************************************************************************************
  UART2 interrupt handler
**************************************************************************************************/
void UART2_Interrupt() iv IVT_INT_USART2 ics ICS_AUTO {
     char R_data;        // Receive data byte
     char T_data;        // Transmit data byte
     BaseType_t xHigherPriorityTaskWoken = pdFALSE;

     if ((RXNE_USART2_SR_bit) /* || (ORE_bit) */ ) {
          R_data = USART2_DR;                // Fetch a byte from UART hardware
          xQueueSendFromISR(U2RcvByteQueueHandle,&R_data,&xHigherPriorityTaskWoken);
     }

     if ((TXE_USART2_SR_bit)&&(TXEIE_USART2_CR1_bit)) {
          T_data = U2XmtBuf[U2XmtBufPtr++];  // Get a byte from the transmit buffer
          if (T_data == NUL) {               // If it is the end of message marker
               TXEIE_USART2_CR1_bit = 0;     // Disable transmitter interrupts
               xSemaphoreGiveFromISR(U2XmtSemHandle,&xHigherPriorityTaskWoken);
          } else {
               USART2_DR = T_data;           // Transfer data byte to transmit register
          }
     }
     portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
}

/**************************************************************************************************
  This routine sets the baud rate.
**************************************************************************************************/
void UART2_SetBaud(void) {

//   Set registers directly
     GPIODEN_bit = 1;             // Set GPIOD clock enable bit

     GPIOD_MODER |= 0x00002800;   // Set port D MODER bits 6 and 5 to 10: alternate function mode
     GPIOD_AFRL |= 0x07700000;    // Select alternate function 0111: AF7 for bits 6 and 5

     USART2EN_bit = 1;            // Set USART2 clock enable bit
     USART2_CR1 |= 0x0000200C;    // Set UE, TE, and RE in USART2 control register 1

     USART2_BRR = 0x00000C35;     // Set baud rate register for 9600 with 30MHz on APB1

/*
     // Could use MikroE library. UART library must be checkmarked
     // Configure UART2 for 9600 baud, 8-bits, no parity, 1 stop bit on pins PD5 (TX) and PD6 (RX)
     UART2_Init_Advanced(9600, _UART_8_BIT_DATA, _UART_NOPARITY, _UART_ONE_STOPBIT, &_GPIO_MODULE_USART2_PD56);
*/
     NVIC_IntEnable(IVT_INT_USART2); // Set interrupt vector
}

/**************************************************************************************************
  This routine initializes 2 queues: a byte queue and a message queue.
**************************************************************************************************/
void UART2_RxQueueInit(void) {
     U2RcvByteQueueHandle = xQueueCreate(U2_RCV_BYTE_Q_LENGTH, sizeof(char));
     vQueueAddToRegistry(U2RcvByteQueueHandle, "U2 RcvByte Queue");
}

/**************************************************************************************************
  This routine starts the receiver.
**************************************************************************************************/
void UART2_StartRx(void) {
     RXNEIE_USART2_CR1_bit = 1;        // Interrupt when receive buffer is full (RXNE=1)
}

/**************************************************************************************************
  UART2 receiver initialization
**************************************************************************************************/
void UART2_RcvInit(void) {
     UART2_SetBaud();
     UART2_RxQueueInit();
     UART2_StartRx();
}

/**************************************************************************************************
  UART_Rx State Machine. Call this function in the main loop.
**************************************************************************************************/
void UART2_Rx(void) {
     U2_RxFuncPtr();    // Use the function pointer to go to a state
}

/**************************************************************************************************
  UART2 receive subroutine
**************************************************************************************************/
void UART2_Rx1(void) {
     char RxChar;

     xQueueReceive(U2RcvByteQueueHandle, &RxChar, 0xFFFFFFFFU); // Block here

     switch (RxChar) {
          case NUL:         // Discard NUL character
               break;

          case '\r':        // Use return character
               /*
               U2RcvBuf[U2RcvBufPtr++] = RxChar;  // Put received character into buffer
               U2RcvBuf[U2RcvBufPtr] = 0;         // End marker
               U2RcvBufPtr = 0;                   // Point to start of buffer
               U2_RxFuncPtr = UART2_Rx2;          // Go to state 2
               */
               break;

          case '\n':        // Use newline character
               //
               U2RcvBuf[U2RcvBufPtr++] = RxChar;  // Put received character into buffer
               U2RcvBuf[U2RcvBufPtr] = 0;         // End marker
               U2RcvBufPtr = 0;                   // Point to start of buffer
               U2_RxFuncPtr = UART2_Rx2;          // Go to state 2
               //
               break;

          default:
               if (U2RcvBufPtr <  sizeof(U2RcvBuf)-2)     // Prevents buffer overrun
               U2RcvBuf[U2RcvBufPtr++] = RxChar;       // Put character in buffer
               break;
     }
}

/**************************************************************************************************
  UART_Rx State 2: Execute the command indicated by the first byte in the buffer then return to
  State 1.
**************************************************************************************************/
void UART2_Rx2(void) {
     char *BufNum;
     char Str[4];
     
     switch (U2RcvBuf[0]) {

          // Change to suit application

          case 'a':  // Diagnostic test for UART2
               UART2_PutQueue("Command 'a' received\r\n");
               break;
          case 'b':
               break;
          case 'c':
               break;
          case 'd':
               break;
          case 'e':
               break;
          case 'f':
               break;
          case 'g':
               break;
          case 'h':
               break;
          case 'i':
               break;
          case 'j':
               break;
          case 'k':
               break;
          case 'l':
               break;
          case 'm':
               break;
          case 'n':
               break;
          case 'o':
               break;
          case 'p':
               break;
          case 'q':
               break;
          case 'r':
               break;
          case 's':
               break;
          case 't':
               break;
          case 'u':
               break;
          case 'v':
               break;
          case 'w':
               break;
          case 'x':
               break;
          case 'y':
               break;
          case 'z': // Force a system reset
               SystemReset();
               break;

          default:
               UART2_PutQueue("Unknown command: ");

               BufNum = GetBuffer();    // Get a free buffer
               (void)sprintf(BufNum,"%c\r\n",U2RcvBuf[0]);
               UART2_PutQueue(BufNum);  // Send the message
               break;
     }
     U2_RxFuncPtr = UART2_Rx1;     // Go back to state 1
}

/**************************************************************************************************
  UART2 receive task
**************************************************************************************************/
//
void UART2_RcvTask(void *pvParameters) {

     UART2_RcvInit();

     while (1) {
          UART2_Rx();    // Check for incoming UART2 messages

          // Delay
          //vTaskDelay(pdMS_TO_TICKS(50)); // Delay for 50ms
     }
}
//
/*************************************************************************************************/










/**************************************************************************************************
  UART2 put queue.
**************************************************************************************************/
void UART2_PutQueue(char *msg) {
     BaseType_t rtnCode;

     rtnCode = xQueueSend(U2XmtPointerQueue, &msg, 0); // No blocking

     if (rtnCode == pdFAIL)
          USB_PutQueue("Failed to enter message pointer into UART2 queue\r\n");
}

/**************************************************************************************************
  This routine starts the transmitter.
**************************************************************************************************/
void UART2_StartTx(void) {
     TXEIE_USART2_CR1_bit = 1;         // Interrupt when transmit buffer is empty (TXE=1)
}

/**************************************************************************************************
  This routine enters a pointer to a text message and starts the transmitter.
**************************************************************************************************/
void UART2_SendMsg(char *s) {
     U2XmtBuf = s;            // Pointer to the message for the interrupt handler
     U2XmtBufPtr = 0;         // Set character counter to start of text message
     UART2_StartTx();         // Start the transmitter
}

/**************************************************************************************************
  This routine checks the transmit buffer queue and if it is not empty, it extracts the pointer
  to the message and sets up the transmitter. Call this function in the main loop.
**************************************************************************************************/
void UART2_Tx(void) {
     char *pcReceivedAddr;    // Last pointer to message obtained from pointer queue
     BaseType_t rtnCode;      // Function return code
     
     xSemaphoreTake(U2XmtSemHandle, portMAX_DELAY); // Block here
     rtnCode = xQueueReceive(U2XmtPointerQueue, &pcReceivedAddr, 0xFFFFFFFFU); // Block here
     
     UART2_SendMsg(pcReceivedAddr);
}

/**************************************************************************************************
  UART2 transmit task
**************************************************************************************************/
//
void UART2_XmtTask(void *pvParameters) {

     U2XmtSemHandle = xSemaphoreCreateBinary(); // Semaphore indicating transmitter is idle
     vQueueAddToRegistry(U2XmtSemHandle, "UART2 transmit Semaphore");
     
     U2XmtPointerQueue = xQueueCreate(64, sizeof(char *)); // Queue of message pointers
     vQueueAddToRegistry(xPointerQueue, "UART2 pointer queue");

     xSemaphoreGive(U2XmtSemHandle); // Initialize semaphore
     UART2_PutQueue("UART2 task started\r\n");
     
     while (1) {
          UART2_Tx();    // Check for outgoing UART2 messages
          
          // Delay
          //vTaskDelay(pdMS_TO_TICKS(50)); // Delay for 50ms
     }
}
//

/**************************************************************************************************
  End of File
**************************************************************************************************/