/**************************************************************************************************
  USB_Driver3.c

  Copyright (C) 2015  Val Gretchev

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************************************
  USB HID Driver for FreeRTOS
  The transmit task blocks when the transmit queue is empty. There was no way to make the receive
  task blocking since the interrupt fires all the time. Therefore, the receive USB task runs all
  the time.

  Development Board: EasyMx PRO v7 for STM32 ARM
  Microcontroller: STM32F407VGT6
  Oscillator: external 25 MHz crystal
  MPU Frequency: 120 MHz from PLL
  
  SW12 bit 7 (VBUS) must be set to the right.

  The USB connector is serviced by the internal USB driver and is configured as USB OTG Full Speed.

  PA11 Pin 70 (USB_OTG_FS_DM)
  PA12 Pin 71 (USB_OTG_FS_DP)
  PA9  Pin 68 (USB_OTG_FS_VBUS)

  The external crystal is 25MHz and HSE is divided by 15 and multiplied by 144 giving a frequency
  of 240MHz. This is further divided by 5 in the Main PLL to produce 48MHz for USB.
  
  On my system, it tales about 11 seconds for USB to inumerate. That is the reason for the
  14-second delay in the start-up function.

**************************************************************************************************/

// Include Files
#include <stdint.h>
#include <stddef.h>
#include "FreeRTOSConfig.h"
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"
#include "timers.h"
#include "projdefs.h"
#include "portmacro.h"

#include "USB_Driver3.h"
#include "Buffers.h"
#include "MyTimers.h"
#include "UART2_driver.h"
#include "RTC.h"


// Buffers
char USB_RdBuff[64];               // USB HID messages are limited to 64 bytes
char USB_WrBuff[64];

// Queues
QueueHandle_t xPointerQueue;       // Queue of pointers

// Semaphores
SemaphoreHandle_t usbSemHandle;    // Used for testing only

// Variables
short USB_Initialized;
TTime DateTimeSet;                 // Date/time structure area


/**************************************************************************************************
   This is the USB interrupt handler
**************************************************************************************************/
void USBInterrupt() iv IVT_INT_OTG_FS ics ICS_AUTO {

     USB_Interrupt_Proc();    // USB interrupt fires all the time. Can't set semaphore here.
}

/**************************************************************************************************
   Hardware Initialization
**************************************************************************************************/
void USB_Initialize(void) {

     HID_Enable(&USB_RdBuff,&USB_WrBuff);
     USB_Initialized = 1;
}

/**************************************************************************************************
   This routine checks USB for a receive message then decodes the first character for a command.
**************************************************************************************************/

// Call this function in main loop
void USB_CheckIn(void) {
     char *BufNum;
     //char i,j;
     //char Ch;
     char Str[4];
     
     //xSemaphoreTake(usbSemHandle,portMAX_DELAY);  // No place to set the semaphore.

     if (HID_Read() != 0) {
          switch (USB_RdBuff[0]) {      // The first character in read buffer is a command
               case 'a': // Diagnostic test of USB
                    USB_PutQueue("Command 'a' received\r\n");
                    break;
               case 'b': // Set date/time from data in message

                    // Send the following: b,YY,MM,DD,DoW(2 digits),HH,MM,SS\r\n
                    // e.g. - b,18,08,04,06,10,23,57\r\n
                    BufNum = GetBuffer();
                    BufNum[0]  = USB_RdBuff[0];   // Command
                    BufNum[1]  = USB_RdBuff[1];   // Comma
                    BufNum[2]  = USB_RdBuff[2];   // Year
                    BufNum[3]  = USB_RdBuff[3];   // Year
                    BufNum[4]  = USB_RdBuff[4];   // Comma
                    BufNum[5]  = USB_RdBuff[5];   // Month
                    BufNum[6]  = USB_RdBuff[6];   // Month
                    BufNum[7]  = USB_RdBuff[7];   // Comma
                    BufNum[8]  = USB_RdBuff[8];   // Day
                    BufNum[9]  = USB_RdBuff[9];   // Day
                    BufNum[10] = USB_RdBuff[10];  // Comma
                    BufNum[11] = USB_RdBuff[11];  // DoW
                    BufNum[12] = USB_RdBuff[12];  // DoW
                    BufNum[13] = USB_RdBuff[13];  // Comma
                    BufNum[14] = USB_RdBuff[14];  // Hour
                    BufNum[15] = USB_RdBuff[15];  // Hour
                    BufNum[16] = USB_RdBuff[16];  // Comma
                    BufNum[17] = USB_RdBuff[17];  // Minute
                    BufNum[18] = USB_RdBuff[18];  // Minute
                    BufNum[19] = USB_RdBuff[19];  // Comma
                    BufNum[20] = USB_RdBuff[20];  // Second
                    BufNum[21] = USB_RdBuff[21];  // Second
                    BufNum[22] = '\r';            // Return
                    BufNum[23] = '\n';            // New line
                    BufNum[24] = 0;               // Nul
                    USB_PutQueue(BufNum);         // Send back date-time for verification

                    //
                    // This code enters the received date-time into RTC registers
                    DateTimeSet.ampm = 0;
                    Str[3] = 0;
                    Str[0] = USB_RdBuff[20];
                    Str[1] = USB_RdBuff[21];
                    DateTimeSet.seconds = atoi(Str);
                    Str[0] = USB_RdBuff[17];
                    Str[1] = USB_RdBuff[18];
                    DateTimeSet.minutes = atoi(Str);
                    Str[0] = USB_RdBuff[14];
                    Str[1] = USB_RdBuff[15];
                    DateTimeSet.hours   = atoi(Str);
                    Str[0] = USB_RdBuff[8];
                    Str[1] = USB_RdBuff[9];
                    DateTimeSet.day     = atoi(Str);
                    Str[0] = USB_RdBuff[5];
                    Str[1] = USB_RdBuff[6];
                    DateTimeSet.month   = atoi(Str);
                    Str[0] = USB_RdBuff[2];
                    Str[1] = USB_RdBuff[3];
                    DateTimeSet.year    = atoi(Str);
                    RTC_Write(&DateTimeSet);

                    UART2_PutQueue("Time and date set in RTC\r\n");
                    //
                    break;
                case 'c': // Show RTC Date/Time
                    GetTime();
                    break;
               case 'd':

                    break;
               case 'e':

                    break;
               case 'f':

                    break;
               case 'g': // Give semaphore
                    if (xSemaphoreGive(usbSemHandle) == pdPASS)
                         USB_PutQueue("Semaphore given\r\n");
                    else
                         USB_PutQueue("Semaphore queue full\r\n");
                    break;
               case 'h': // Take semaphore
                    if (xSemaphoreTake(usbSemHandle,0) == pdPASS)
                         USB_PutQueue("Had a semaphore\r\n");
                    else
                         USB_PutQueue("Semaphore queue empty\r\n");
                    break;
               case 'i':
                    break;
               case 'j': // Start auto-reload timer with this command
                    StartAutoReloadTimer();
                    break;
               case 'k': // Test to turn PD8 LED on
                    IndOn();
                    break;
               case 'l': // Test to turn PD8 LED off
                    IndOff();
                    break;
               case 'm': // Test to toggle PD8 LED
                    IndTog();
                    break;
               case 'n': // Change auto-reload timer period to 100ms. Also starts the timer.
                    ChangeReloadTimerPeriod(100);
                    break;
               case 'o': // Stop auto-reload timer
                    StopAutoReloadTimer();
                    break;
               case 'p': // Display the number of free bytes in the heap
                    BufNum = GetBuffer();
                    sprintf(BufNum,"%Ld\r\n",xPortGetFreeHeapSize());
                    USB_PutQueue(BufNum);
                    break;
               case 'q': // Display the minimum number of bytes on the heap since starting
                    BufNum = GetBuffer();
                    sprintf(BufNum,"%Ld\r\n",xPortGetMinimumEverFreeHeapSize());
                    USB_PutQueue(BufNum);
                    break;
               case 'r': //
                    break;
               case 's':
                    UART2_PutQueue("Hello\r\n");
                    break;
               case 't': // This command starts the one-shot timer
                    StartOneShotTimer();
                    break;
               case 'u': // Change timer period to 1000ms. Also starts the timer.
                    ChangeTimerPeriod(1000);
                    break;
               case 'v': // Change timer period to 100ms. Also starts the timer.
                    ChangeTimerPeriod(100);
                    break;
               case 'w': // Test for hardware timer
                    Timer8_Ch2_Start(); // Flash PC7 LED
                    break;
               case 'x': // Send many messages to USB
                    DisplayTIM8Regs();
                    break;
               case 'y':
                    break;

               case 'z': // Force a system reset
                    SystemReset();
                    break;

                    // Etc.

               default:  // Catch all other unimplemented commands
                    BufNum = GetBuffer();  // Get a free buffer
                    (void)sprintf(BufNum,"Unknown command: %c\r\n",USB_RdBuff[0]);
                    USB_PutQueue(BufNum);
                    break;
          }
     }
}

/**************************************************************************************************
   This routine makes an entry into the transmit queue FIFO.
**************************************************************************************************/
void USB_PutQueue(char *msg) {
     BaseType_t rtnCode;

     rtnCode = xQueueSend(xPointerQueue, &msg, 0); // No blocking

     if (rtnCode == pdFAIL)
          Timer8_Ch2_Start(); // Flash PC7 LED if pdFAIL
}

/**************************************************************************************************
   This routine waits for a pointer from the transmit queue FIFO, then sends the message to USB.
**************************************************************************************************/

// Call this function in main loop
void USB_CheckOut(void) {
     char RetryCount = 0;          // Set to zero on entry to this function
     BaseType_t rtnCode;           // Function return code
     char *pcReceivedAddr;         // Last pointer to message obtained from pointer queue

     rtnCode = xQueueReceive(xPointerQueue, &pcReceivedAddr, 0xFFFFFFFFU); // Block here

     if (rtnCode == pdPASS) {
          while (((HID_Write(pcReceivedAddr,64) == 0)) && (RetryCount < 5)) { // Send 64 bytes to USB
               RetryCount++;       // Try 5 times
          }
     } else {
          Timer8_Ch2_Start();      // Flash PC7 LED if pdFAIL
     }
}

/**************************************************************************************************
  USB receive driver task
**************************************************************************************************/
//
void usbRcvTask(void *pvParameters) {

     USB_Initialize();             // Initialize USB hardware

     // This is unused for now
     usbSemHandle = xSemaphoreCreateBinary();
     vQueueAddToRegistry(usbSemHandle, "USB receive Semaphore");

     while (1) {
          USB_CheckIn(); // Check for incoming USB messages
          //vTaskDelay(pdMS_TO_TICKS(100));
     }
}
//

/**************************************************************************************************
  USB trasmit driver task
**************************************************************************************************/
//
void usbXmtTask(void *pvParameters) {

     // Create a queue of pointers
     xPointerQueue = xQueueCreate(64, sizeof(char *));
     vQueueAddToRegistry(xPointerQueue, "USB message pointer queue");

     vTaskDelay(pdMS_TO_TICKS(14000));
     USB_PutQueue(" ");                      // Output blank line
     USB_PutQueue("Program started\r\n");    // Send opening message on USB
     
     // Start diagnostic timer for PC7 LED
     Timer8_Initialize();
     Timer8_Ch2_Start();

     while (1) {
          USB_CheckOut(); // Check for outgoing USB messages
     }
}
//


/**************************************************************************************************
  End of File
**************************************************************************************************/